from aiohttp.web import middleware
from typing import TYPE_CHECKING

from aiohttp_sqlalchemy.constants import DEFAULT_KEY
from aiohttp_sqlalchemy.exceptions import DuplicateRequestKeyError

if TYPE_CHECKING:
    from aiohttp.web import Request, StreamResponse
    from typing import Callable


def sa_middleware(key: str = DEFAULT_KEY) -> 'Callable':
    """ SQLAlchemy asynchronous middleware factory. """
    @middleware
    async def sa_middleware_(request: 'Request', handler: 'Callable')\
            -> 'StreamResponse':
        if key in request:
            raise DuplicateRequestKeyError(key)

        Session = request.config_dict.get(key)
        async with Session() as request[key]:
            return await handler(request)

    return sa_middleware_



from sqlalchemy.ext.asyncio import AsyncSession
from typing import TYPE_CHECKING

from aiohttp_sqlalchemy.constants import DEFAULT_KEY
from aiohttp_sqlalchemy.decorators import sa_decorator
from aiohttp_sqlalchemy.exceptions import DuplicateAppKeyError, DuplicateRequestKeyError
from aiohttp_sqlalchemy.middlewares import sa_middleware
from aiohttp_sqlalchemy.views import SABaseView, SAView


if TYPE_CHECKING:
    from aiohttp.web import Application
    from typing import Callable, Iterable, Tuple

    TSessionFactory = Callable[..., AsyncSession]
    TSABinding = Tuple[TSessionFactory, str, bool]


__version__ = '0.9.0'

__all__ = ['DuplicateAppKeyError', 'DuplicateRequestKeyError', 'SABaseView',
           'sa_bind', 'sa_decorator', 'sa_middleware', 'SAView', 'setup',]


def sa_bind(factory: 'TSessionFactory', key: str = DEFAULT_KEY, *,
            middleware: bool = True) -> 'TSABinding':
    """ Session factory wrapper for binding in setup function. """
    return factory, key, middleware


def setup(app: 'Application', bindings: 'Iterable[TSABinding]'):
    """ Setup function for binding SQLAlchemy engines. """
    for Session, key, middleware in bindings:
        if key in app:
            raise DuplicateAppKeyError(key)

        app[key] = Session

        if middleware:
            app.middlewares.append(sa_middleware(key))



class AbstractDuplicateKeyError(ValueError):
    message: str


class DuplicateAppKeyError(AbstractDuplicateKeyError):
    def __init__(self, key: str):
        self.message = f'Duplicated app key `{key}`. Check `engines` ' \
                       f' argument in `aiohttp_sqlalchemy.setup()` call.'
        super().__init__(self.message)


class DuplicateRequestKeyError(AbstractDuplicateKeyError):
    def __init__(self, key: str):
        self.message = f'Duplicated request key `{key}`. Check middlewares ' \
                       f'and decorators from `aiohttp_sqlalchemy`.'
        super().__init__(self.message)



from abc import ABCMeta
from aiohttp.abc import AbstractView
from aiohttp.web import View
from typing import TYPE_CHECKING

from aiohttp_sqlalchemy.constants import DEFAULT_KEY

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession
    from typing import Any


class SAAbstractView(AbstractView, metaclass=ABCMeta):
    """
    Simple SQLAlchemy view based on aiohttp.abc.AbstractView.

    The `__await__` method must be implemented in child classes.

    Suitable for a specific usage with multiple models.
    """
    def sa_session(self, key: str = DEFAULT_KEY) -> 'AsyncSession':
        return self.request[key]


class SAOneModelMixin(SAAbstractView, metaclass=ABCMeta):
    """
    One model SQLAlchemy view based on aiohttp.abc.AbstractView.

    The `__await__` method must be implemented in child classes.

    Suitable for a usually usage with one model.
    """
    sa_model: 'Any'  # Not all developers use declarative mapping


class SABaseView(View, SAAbstractView):
    """
    Simple SQLAlchemy view based on aiohttp.web.View.

    Recomended for a specific usage with multiple models.
    """


class SAView(View, SAOneModelMixin):
    """
    One model SQLAlchemy view based on aiohttp.web.View.

    Recomended for a usually usage with one model.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiohttp_sqlalchemy']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.17,<2.0.0', 'aiohttp>=3.7.4.post0,<4.0.0']

setup_kwargs = {
    'name': 'aiohttp-sqlalchemy',
    'version': '0.9.0',
    'description': 'SQLAlchemy 1.4 / 2.0 support for aiohttp.',
    'long_description': "==================\naiohttp-sqlalchemy\n==================\nSQLAlchemy 1.4 / 2.0 support for aiohttp.\n\nLibrary forward a ``sqlalchemy.ext.asyncio.AsyncSession`` object as\n``request['sa_main']`` or ``SAView.sa_session()`` by default.\n\n\nDocumentation\n-------------\nhttps://aiohttp-sqlalchemy.readthedocs.io/\n\n\nInstallation\n------------\n::\n\n    pip install aiohttp-sqlalchemy\n\n\nSimple example\n--------------\nInstall ``aiosqlite`` for work with sqlite3: ::\n\n  pip install aiosqlite\n\nCopy and paste this code in a file and run:\n\n.. code-block:: python\n\n  from aiohttp import web\n  import aiohttp_sqlalchemy\n  from aiohttp_sqlalchemy import sa_bind\n  from datetime import datetime\n  import sqlalchemy as sa\n  from sqlalchemy import orm\n  from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine\n\n\n  metadata = sa.MetaData()\n  Base = orm.declarative_base(metadata=metadata)\n\n\n  class MyModel(Base):\n      __tablename__ = 'my_table'\n      id = sa.Column(sa.Integer, primary_key=True)\n      timestamp = sa.Column(sa.DateTime(), default=datetime.now)\n\n\n  async def main(request):\n      async with request['sa_main'].bind.begin() as conn:\n          await conn.run_sync(Base.metadata.create_all)\n\n      async with request['sa_main'].begin():\n          request['sa_main'].add_all([MyModel()])\n          result = await request['sa_main'].execute(sa.select(MyModel))\n          data = {r.id: r.timestamp.isoformat() for r in result.scalars()}\n          return web.json_response(data)\n\n\n  app = web.Application()\n\n  engine = create_async_engine('sqlite+aiosqlite:///')\n  Session = orm.sessionmaker(engine, AsyncSession)\n  aiohttp_sqlalchemy.setup(app, [sa_bind(Session)])\n\n  app.add_routes([web.get('/', main)])\n\n  if __name__ == '__main__':\n      web.run_app(app)\n",
    'author': 'Ruslan Ilyasovich Gilfanov',
    'author_email': 'ri.gilfanov@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ri-gilfanov/aiohttp-sqlalchemy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from aiohttp.abc import AbstractView
from functools import wraps
from typing import TYPE_CHECKING

from aiohttp_sqlalchemy.constants import DEFAULT_KEY
from aiohttp_sqlalchemy.exceptions import DuplicateRequestKeyError

if TYPE_CHECKING:
    from aiohttp.web import StreamResponse


def sa_decorator(key: str = DEFAULT_KEY):
    """ SQLAlchemy asynchronous handler decorator. """
    def wrapper(handler):
        @wraps(handler)
        async def wrapped(*args, **kwargs) -> 'StreamResponse':
            print(*args)
            request = args[0].request \
                      if isinstance(args[0], AbstractView) \
                      else args[-1]

            if key in request:
                raise DuplicateRequestKeyError(key)

            Session = request.config_dict.get(key)
            async with Session() as request[key]:
                return await handler(*args, **kwargs)

        return wrapped
    return wrapper
